@@ -1,4 +1,5 @@
 import asyncio
+import os
 from enum import Enum
 from typing import TYPE_CHECKING
 
@@ -16,9 +17,6 @@ from machine_learning.recommending.movielens.data import MovieLens25m
 if TYPE_CHECKING:
     from machine_learning.recommending.interface import RecommenderModuleBase
 
-state_dict_path = "/Users/dimakoshman/MovieRecommender/svd.pt"
-movie_lens_25m_directory = "/Users/dimakoshman/MovieRecommender/ml-25m"
-
 
 class Feedback(Enum):
     NONE = "Haven't seen it"
@@ -29,12 +27,19 @@ class Feedback(Enum):
         return {v: i for i, v in enumerate(Feedback)}[self]
 
 
+def load_model(torch_nn_module, state_dict_path):
+    state_dict = torch.load(state_dict_path)
+    model = torch_nn_module()
+    model.load_state_dict(state_dict)
+    return model
+
+
 class MovieMarkdownGenerator:
     def __init__(
         self,
         links_dataframe,
         movies_dataframe,
-        tmdb_api_token="0369096d5891e7959f8e5ae35bae71bc",
+        tmdb_api_token,
     ):
         self.links = links_dataframe
         self.movies_dataframe = movies_dataframe
@@ -93,14 +98,14 @@ class MovieMarkdownGenerator:
         return markdown, poster_url
 
 
-class AsyncRecommender:
+class AheadRecommender:
     def __init__(self, recommender: "RecommenderModuleBase"):
         self.recommender = recommender
         self.user_explicit = pd.Series(
             data=np.full(self.recommender.n_items, fill_value=np.nan)
         )
         self.movie_id = self.recommend()
-        self.task = asyncio.create_task(self.async_recommend())
+        self.next_movie_id = None
 
     def recommend(self):
         user_explicit = scipy.sparse.coo_matrix(
@@ -114,21 +119,25 @@ class AsyncRecommender:
         movie_id = user_recommendations[new_items_positions[0]]
         return movie_id
 
-    async def async_recommend(self):
-        self.user_explicit.loc[self.movie_id] = 0
-        return self.recommend()
-
-    async def next_recommendation(self, rating_of_previous_movie: int):
-        wandb.log(
-            dict(
-                movie_id=self.movie_id,
-                rating=rating_of_previous_movie,
+    def next_recommendation(self):
+        if self.next_movie_id is not None:
+            raise ValueError(
+                "Before generating another recommendation, "
+                "rating for previous recommendation must be saved."
             )
-        )
-        self.user_explicit.loc[self.movie_id] = rating_of_previous_movie
-        self.movie_id = await self.task
-        self.task = asyncio.create_task(self.async_recommend())
-        return self.movie_id
+        self.user_explicit.loc[self.movie_id] = 0
+        self.next_movie_id = self.recommend()
+        return self.next_movie_id
+
+    def save_rating(self, movie_id: int, rating: int):
+        if movie_id != self.movie_id:
+            raise ValueError(
+                f"Provided movie id {movie_id} doesn't "
+                f"match stored movie id {self.movie_id}"
+            )
+        wandb.log(dict(movie_id=self.movie_id, rating=rating))
+        self.user_explicit.loc[self.movie_id] = rating
+        self.movie_id, self.next_movie_id = self.next_movie_id, None
 
     def __del__(self):
         wandb.log(
@@ -136,42 +145,81 @@ class AsyncRecommender:
         )
 
 
-state_dict = torch.load(state_dict_path)
-model = SVDRecommender()
-model.load_state_dict(state_dict)
-model.eval()
-movielens = MovieLens25m(movie_lens_25m_directory)
-movie_markdown_generator = MovieMarkdownGenerator(
-    links_dataframe=movielens["links"],
-    movies_dataframe=movielens["movies"],
-)
+class AsyncRecommendingAppContentDispenser:
+    def __init__(self, recommender: "RecommenderModuleBase", movie_markdown_generator):
+        self.ahead_recommender = AheadRecommender(recommender=recommender)
+        self.movie_markdown_generator = movie_markdown_generator
+        self.movie_id = self.ahead_recommender.movie_id
+        self.initial_content = self.movie_markdown_generator(self.movie_id)
+        self.next_content = asyncio.create_task(self.content_ahead())
 
-with gr.Blocks() as app:
-    async_recommender = AsyncRecommender(recommender=model)
+    async def content_ahead(self):
+        movie_id = self.ahead_recommender.next_recommendation()
+        content = self.movie_markdown_generator(movie_id)
+        return movie_id, content
+
+    async def content(self, feedback: Feedback):
+        self.ahead_recommender.save_rating(movie_id=self.movie_id, rating=feedback.int)
+        self.movie_id, content = await self.next_content
+        self.next_content = asyncio.create_task(self.content_ahead())
+        return content
+
+
+def build_app_blocks(
+    recommender: "RecommenderModuleBase", movie_markdown_generator
+) -> None:
+    dispenser = AsyncRecommendingAppContentDispenser(
+        recommender=recommender,
+        movie_markdown_generator=movie_markdown_generator,
+    )
     gr.Markdown("This movie recommender adapts to your preferences")
     with gr.Row():
-        first_movie_markdown, first_image_url = movie_markdown_generator(
-            async_recommender.movie_id
-        )
+        first_movie_markdown, first_image_url = dispenser.initial_content
         image = gr.Image(first_image_url, interactive=False)
         image.style(width=200)
         with gr.Column(scale=4):
             markdown = gr.Markdown(first_movie_markdown)
 
-            def make_button(feedback):
+            def build_button(feedback):
                 async def wrapper():
-                    movie_id = await async_recommender.next_recommendation(feedback.int)
-                    movie_markdown, image_url = movie_markdown_generator(movie_id)
-                    return movie_markdown, image_url
+                    return await dispenser.content(feedback=feedback)
 
                 button = gr.Button(feedback.value)
-                button.click(
-                    fn=wrapper,
-                    inputs=[],
-                    outputs=[markdown, image],
-                )
+                button.click(fn=wrapper, inputs=[], outputs=[markdown, image])
                 return button
 
-            none_button = make_button(Feedback.NONE)
+            build_button(Feedback.NONE)
             with gr.Row():
-                buttons = [make_button(feedback) for feedback in list(Feedback)[1:]]
+                for rated_feedback in list(Feedback)[1:]:
+                    build_button(rated_feedback)
+
+
+def main():
+    project = "Recommending"
+    entity = "dkoshman"
+    path_to_movielens_folder = "local/ml-25m"
+    state_dict_path = "svd.pt"
+    tmdb_api_token = os.environ["TMDB_API_TOKEN"]
+
+    movielens = MovieLens25m(path_to_movielens_folder=path_to_movielens_folder)
+    movie_markdown_generator = MovieMarkdownGenerator(
+        links_dataframe=movielens["links"],
+        movies_dataframe=movielens["movies"],
+        tmdb_api_token=tmdb_api_token,
+    )
+    model = load_model(torch_nn_module=SVDRecommender, state_dict_path=state_dict_path)
+    model.eval()
+
+    with gr.Blocks() as app:
+        build_app_blocks(
+            recommender=model,
+            movie_markdown_generator=movie_markdown_generator,
+        )
+
+    wandb.finish()
+    wandb.init(job_type="app", project=project, entity=entity)
+    app.launch()
+
+
+if __name__ == "__main__":
+    main()
